/* $Header: /src/carson/nistnet/RCS/linmunge.c,v 1.1 1997/10/20 17:11:39 carson Exp $ */

/* linmunge.c - Sample add-on packet "munger."
 *
 * This code exists as a loadable kernel module.  It requires the nistnet
 * emulator to have been previously installed.  It also provides a device
 * driver interface, which allows user-level access to the packet munging
 * routines.
 *
 * For the sake of simplicity, many basic routines are patterned after
 * (swiped from) corresponding ones in knistnet.c.
 *
 * Principles of operation: open the device, then use the MUNGEIOCTL_SET
 * ioctl to define which source/address pair you wish to monitor.  Operations
 * then available are:
 *      MUNGEIOCTL_WAIT - packets are held until read/examined
 *      MUNGEIOCTL_NOWAIT - stats are recorded, but packet not delayed (default)
 *      MUNGEIOCTL_STAT - get stat info on most recent packet (and release if
 *                        WAIT
 *      read()          - get contents of current packet (only if WAIT)
 *
 * read and STAT operations will be blocking unless O_NONBLOCK has been
 * set.  Note this affects whether the read or stat will wait.  Setting
 * whether the packet will wait is done with the ioctls above.
 *
 * select may be used to see whether a packet or information is available.
 *
 * To associate this information with a particular file descriptor, I use
 * the private_data area in the file structure.  After all, the comment
 * in the header says "needed for tty driver, and maybe others."  So
 * there's a definite "maybe" saying I can use it!
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/version.h>

#include <linux/fs.h>
#include <linux/major.h>
#include <asm/segment.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/param.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/interrupt.h>
#include <linux/signal.h>
#include <linux/malloc.h>

#include <asm/io.h>
#include <asm/irq.h>

#include <linux/mc146818rtc.h>
#include <linux/timex.h>
#include <linux/config.h>

#include <linux/socket.h>
#include <linux/in.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <net/ip.h>

#include "tabledist.h"

#include "linux/fast_timer.h"

#include "nistnet.h"

#ifndef MODULE
#define BREAKPOINT() asm("   int $3")
#else
#define BREAKPOINT()
#endif

void onmunge(void), offmunge(void);
void packetwait(struct lin_mungebox *mungereq, struct sk_buff *skb, struct device *dev, struct packet_type *ptype, struct lin_hitbox *hitme);
void packetwake(struct lin_mungebox *mungereq);
int readerwait(struct lin_mungebox *mungereq);
void readerwake(struct lin_mungebox *mungereq);
static int munge_debug = 0;

#define HASHSIZE 256
static struct lin_mungebox *mungetable[HASHSIZE] = {0};

void
freemungetable(void)
{
	struct lin_mungebox *freeme, *frome;
	int i;

	for (i=0; i < HASHSIZE; ++i) {
		freeme = mungetable[i];
		while (freeme) {
			frome = freeme->next;
			kfree_s(freeme, sizeof(struct lin_mungebox));
			freeme = frome;
		}
	}
}


/* Hash an entry by src/dest address */
int
mungehash(__u32 src, __u32 dest)
{
	unsigned int value, answer=0;

	/* this is not intended to be an extraordinarily good hash! */
	value = (unsigned int)src*69069 + (unsigned int)dest;
	while (value) {
		answer |= value;
		value >>= 8;
	}
	return (int) (answer&(HASHSIZE-1));
}

struct lin_mungebox *
findmunge(__u32 src, __u32 dest)
{
	register struct lin_mungebox *answer;
	
	answer = mungetable[mungehash(src, dest)];
	while (answer) {
		if (answer->munge_who.src == src
			&& answer->munge_who.dest == dest) {
			return answer;
		}
		answer = answer->next;
	}

	/* Try again with generic addresses */
	answer = mungetable[mungehash(src, INADDR_ANY)];
	while (answer) {
		if (answer->munge_who.src == src
			&& answer->munge_who.dest == INADDR_ANY) {
			return answer;
		}
		answer = answer->next;
	}

	answer = mungetable[mungehash(INADDR_ANY, dest)];
	while (answer) {
		if (answer->munge_who.src == INADDR_ANY
			&& answer->munge_who.dest == dest) {
			return answer;
		}
		answer = answer->next;
	}
	answer = mungetable[mungehash(INADDR_ANY, INADDR_ANY)];
	while (answer) {
		if (answer->munge_who.src == INADDR_ANY
			&& answer->munge_who.dest == INADDR_ANY) {
			return answer;
		}
		answer = answer->next;
	}
	return NULL;
}

int
rmmungereq(struct lin_mungebox *mungereq)
{
	struct lin_mungebox *victim, *parent;
	
	victim = mungetable[mungehash(mungereq->munge_who.src, mungereq->munge_who.dest)];
	if (!victim) return -ESRCH;
	if (victim == mungereq) {
		mungetable[mungehash(mungereq->munge_who.src, mungereq->munge_who.dest)]
					= victim->next;
		kfree_s(victim, sizeof(struct lin_mungebox));
	} else for (parent=victim, victim=victim->next; victim;
			parent=victim, victim=victim->next) {
		if (victim == mungereq) {
			parent->next = victim->next;
			kfree_s(victim, sizeof(struct lin_mungebox));
			break;
		}
	}
	return 0;
}


int
addmungereq(struct lin_mungebox *mungereq)
{
	/* Initialize fields */
	mungereq->munge_flags = 0;
	memset(&mungereq->munge_stats, 0, sizeof(struct lin_mungestats));
	mungereq->reader_wait = NULL;
	mungereq->packet_wait = NULL;
	/* Insert in table */
	mungereq->next = mungetable[mungehash(mungereq->munge_who.src, mungereq->munge_who.dest)];
	mungetable[mungehash(mungereq->munge_who.src, mungereq->munge_who.dest)] = mungereq;
	return 0;
}


/* nice and high, but it is tunable: insmod mungemod.o major=your_selection */
static int major = 63;	/* .0625 = 1/16, Lina Inverse is 16 in NEXT */

/*
 * The driver.
 */

/* read -- should read packet contents, but we'll do header for now */
static int
mw_read(struct inode * node,struct file * file,char * buf,int count)
{
	struct lin_mungebox *readme=NULL;

	if (file) {
		readme = (struct lin_mungebox *)file->private_data;
	}
	if (!readme)
		return -ENODEV;	/* ha, ha */
	if (count > sizeof(struct lin_mungestats))
		count = sizeof(struct lin_mungestats);
	memcpy_tofs(buf, (char *)&readme->munge_stats, (unsigned long) count);
	return count;
}

static int
mw_write(struct inode * node,struct file * file,const char * buf,int count)
{
	/* Will allow packet writing... */
	return 0;
}

/*
 * mucky muck ioctl interface
 */
static int
mw_ioctl(struct inode * inode, struct file * file, unsigned int type, unsigned long arg)
{
	struct lin_mungebox *mungereq=NULL;
	int rc = 0;

	if (file)
		mungereq = (struct lin_mungebox *)file->private_data;
	switch (type) {
	case MUNGEIOCTL_OFF:
		/* Shut down */
		offmunge();
		break;
	case MUNGEIOCTL_ON:
		/* Turn on if not already on */
		onmunge();
		break;
	case MUNGEIOCTL_WAIT:
		/* switch on packet waiting */
		if (mungereq)
			mungereq->munge_flags |= MUNGE_WAIT;
		else
			rc = -ESRCH;
		break;
	case MUNGEIOCTL_NOWAIT:
		/* switch off packet waiting */
		/* Turn on if not already on */
		if (mungereq)
			mungereq->munge_flags &= ~MUNGE_WAIT;
		else
			rc = -ESRCH;
		break;
	case MUNGEIOCTL_SET:
		/* Get what they want */
		if (mungereq)
			kfree_s(mungereq, sizeof(struct lin_mungebox));
		mungereq = (struct lin_mungebox *)kmalloc(sizeof(struct lin_mungebox), GFP_KERNEL);
		if (!mungereq)
			rc = -ENOMEM;
		memcpy_fromfs(&mungereq->munge_who, (struct addrpair *)arg,
			sizeof(struct addrpair));
		/* Add it to the table */
		rc = addmungereq(mungereq);
		/* Add it to the file pointer */
		file->private_data = (void *)mungereq;
		break;
	case MUNGEIOCTL_UNSET:
		/* Remove it from the table */
		if (mungereq)
			rc = rmmungereq(mungereq);
		else
			rc = -ESRCH;
		/* Clear out file pointer */
		file->private_data = NULL;
		break;
	case MUNGEIOCTL_STAT:
		/* Copy out individual stats */
		if (mungereq) {
			/* If waiting, check whether new data is available */
			if (!(file->f_flags & O_NONBLOCK))
				rc = readerwait(mungereq);
			if (!rc)
				memcpy_tofs((struct lin_mungestats *)arg,
					&mungereq->munge_stats,
					sizeof(struct lin_mungestats));
			/* @@ clear data read flag */
			/* @@ yes, minor race condition, I suppose, but
			 * supporting multiple readers is not so important
			 * ... */
			mungereq->munge_flags &= ~MUNGE_UPDATED;

			/* allow packet to proceed if waiting */
			packetwake(mungereq);
		}
		else
			rc = -ESRCH;
		break;
	case MUNGEIOCTL_DEBUG:
		munge_debug = get_user((long *)arg);
		break;
	}
	return rc;
}


int
mw_select(struct inode *inode, struct file *file, int sel_type, select_table *wait)
{
	struct lin_mungebox *mungereq=NULL;

	if (sel_type != SEL_IN)
		return 0;
	if (file) {
		mungereq = (struct lin_mungebox *)file->private_data;
	} else
		return 0;
	if (mungereq->munge_flags & MUNGE_UPDATED)
		return 1;
	select_wait(&mungereq->reader_wait, wait);
	return 0;
}
/*
 * Our special open code.
 * MOD_INC_USE_COUNT make sure that the driver memory is not freed
 * while the device is in use.
 */
static int
mw_open( struct inode *inode, struct file *file)
{
	MOD_INC_USE_COUNT;
	if (file) {
		file->private_data = NULL;
	}
	return 0;   
}

/*
 * Now decrement the use count.
 */
static void
mw_close( struct inode* ino, struct file* file)
{
	struct lin_mungebox *mungereq=NULL;

	if (file) {
		mungereq = (struct lin_mungebox *)file->private_data;
		if (mungereq)
			(void) rmmungereq(mungereq);
		file->private_data = NULL;
	}
	MOD_DEC_USE_COUNT;
}

static struct file_operations mw_fops = {
	NULL,
	mw_read,
	mw_write,
	NULL,		/* mw_readdir */
	mw_select,
	mw_ioctl,
	NULL,
	mw_open,
	mw_close,
	NULL		/* fsync */
};

void
runpacket(struct sk_buff *skb, struct device *dev, struct packet_type *ptype, struct lin_hitbox *hitme)
{
	(void) ptype->func(skb, dev, ptype);
}

void
packetwake(struct lin_mungebox *mungereq)
{

}
void
packetwait(struct lin_mungebox *mungereq, struct sk_buff *skb, struct device *dev, struct packet_type *ptype, struct lin_hitbox *hitme)
{
	/* @@ wait until read/stat.... */
	runpacket(skb, dev, ptype, hitme);
}


/* Check if new data is available.  If not, wait for it. */
int
readerwait(struct lin_mungebox *mungereq)
{
	unsigned long flags;
	struct wait_queue wait = {current, NULL};

	/* Other code doesn't seem to do this, but isn't there an
	 * obvious race condition otherwise?
	 */
	save_flags(flags);
	cli();
	if (!(mungereq->munge_flags & MUNGE_UPDATED)) {
		/* Since we've already blocked interrupts, we can
		 * call __add_wait_queue and __remove_wait_queue
		 * directly.
		 */
		__add_wait_queue(&mungereq->reader_wait, &wait);
		interruptible_sleep_on(&mungereq->reader_wait);
		__remove_wait_queue(&mungereq->reader_wait, &wait);
	};
	restore_flags(flags);
	/* @@ Check whether we were interrupted */
	if (current->signal & ~current->blocked) {
		return -EINTR;	/* @@ right rc? */
	} else if (!(mungereq->munge_flags & MUNGE_UPDATED)) {
		/* @@ what now? try again? */
		return 0;
	} else {
		return 0;
	}

}

void
readerwake(struct lin_mungebox *mungereq)
{
	/* Mark this as a new packet */
	mungereq->munge_flags |= MUNGE_UPDATED;
	/* Wake up those waiting */
	wake_up_interruptible(&mungereq->reader_wait);
}

/* 
 * return -1 for error/drop
 * return 0 for "we process" (e.g. delay)
 * return 1 for "you process"
 */
int
linmunge(struct sk_buff *skb, struct device *dev, struct packet_type *ptype, struct lin_hitbox *hitme)
{
	struct lin_mungebox *mungeme;
	struct iphdr *iph;
	unsigned char *raw;

	/* Get the ip header */
	iph = skb->h.iph;
	mungeme = findmunge(iph->saddr, iph->daddr);
	if (!mungeme)
		return 1;
	/* @@ ulp, should get options also... */
	mungeme->munge_stats.ip = *iph;
	/* Now we want to find the part past the header */
	raw = skb->h.raw + iph->ihl*4;
	switch (iph->protocol) {
	case IPPROTO_IPIP:
		break;
	case IPPROTO_TCP:
		{
		struct tcphdr *th;

		th = (struct tcphdr *)raw;
		mungeme->munge_stats.h.tcp = *th;
		break;
		}
	case IPPROTO_UDP:
		{
		struct udphdr *uh;

		uh = (struct udphdr *)raw;
		mungeme->munge_stats.h.udp = *uh;
		break;
		}
	case IPPROTO_ICMP:
		{
		struct icmphdr *ih;

		ih = (struct icmphdr *)raw;
		mungeme->munge_stats.h.icmp = *ih;
		break;
		}
	case IPPROTO_IGMP:
		{
		struct igmphdr *ih;

		ih = (struct igmphdr *)raw;
		mungeme->munge_stats.h.igmp = *ih;
		break;
		}
	}
	if (mungeme->munge_flags & MUNGE_WAIT) {
		/* wait and run packet*/
		packetwait(mungeme, skb, dev, ptype, hitme);
		return 0;
	}
	/* Wake up anyone waiting */
	readerwake(mungeme);

	return 1;
}

static int doingmunge = 0;

void
onmunge(void)
{
	if (!doingmunge) {
		++doingmunge;
		addmunge(linmunge);
printk("munge: on\n");
	}
}

void
offmunge(void)
{
	if (doingmunge) {
		doingmunge=0;
		rmmunge(linmunge);
printk("munge: off\n");
	}
}
#ifdef MODULE
int
init_module( void)
#else
int
linmunge_init(void)
#endif
{

	printk( "linmunge: init_module called\n");
	if (register_chrdev(major, "mw", &mw_fops)) {
		printk("linmunge: register_chrdev failed: goodbye world :-(\n");
		return -EIO;
	}
	else
		printk("linmunge: Hello, world\n");

	return 0;
}

#ifdef MODULE
void
cleanup_module(void)
{
	printk("linmunge:  cleanup_module called\n");

	offmunge();
	if (unregister_chrdev(major, "mw") != 0)
		printk("linmunge: cleanup_module failed\n");
	else
		printk("linmunge: cleanup_module succeeded\n");
}
#endif
